import colorama
import sys
colorama.init()

def print_error(err_message):
    print(f"{colorama.Fore.LIGHTRED_EX}   ascvid: !! [ERROR] !!\n{err_message}{colorama.Fore.RESET}",file=sys.stderr)
def print_warning(warning,ask=None):
    print(f"{colorama.Fore.LIGHTYELLOW_EX}   ascvid: ! [WARNING] !\n{warning}{colorama.Fore.RESET}",file=sys.stderr)
    if ask is not None:
        print(f"{colorama.Fore.LIGHTCYAN_EX} ascvid: ? [DIALOG] ? => {ask} ?{colorama.Fore.RESET}",file=sys.stderr,end=None)
        return input()




import time

class Timer:
    def __init__(self):
        self._start_time=None
        self.paused=True
        self.__running_total=0
    @property
    def curtime(self):
        if self._start_time is None:
            raise RuntimeError
        if self.paused:
            return self.__running_total
        return self.__running_total+(time.perf_counter()-self._start_time)
    def pause(self):
        if not self.paused:
            self.__running_total=self.curtime
            self.paused=True
    def start(self):
        if self.paused:
            self._start_time=time.perf_counter()
            self.paused=False
            



import sounddevice as sd
import time
import moviepy.editor as mp
from .timer import Timer
class Audio:
    def __init__(self,clip):
        self.clip=clip
        self.audio=clip.to_soundarray()
        self.fps=clip.fps
        self.start_time=-1
        self.resume_time=0
        self._playing=False
        self.timer=Timer()
    def play(self):
        self.timer.start()
        sd.play(self.audio,self.fps)
    @property
    def playing(self):
        if not self._playing:
            return False
        return self._playing and (not sd.get_stream().stopped)
    def pause(self):
        self.timer.pause()
        self._playing=False
        sd.stop()
    def resume(self):
        
        a=self.audio[int(self.fps*self.timer.curtime):]        
        self.timer.start()
        sd.play(a,self.fps)

    



import colorama
from math import sqrt
import colorsys
colorama.init()
#### Please note that these colors might display differently on other terminals
_COLORS = {
    (0, 0, 0): colorama.Fore.BLACK,
    (128, 0, 0): colorama.Fore.RED,
    (0, 128, 0): colorama.Fore.GREEN,
    (128, 128, 0): colorama.Fore.YELLOW,
    (0, 0, 128): colorama.Fore.BLUE,
    (128, 0, 128): colorama.Fore.MAGENTA,
    (0, 128, 128): colorama.Fore.CYAN,
    (192, 192, 192): colorama.Fore.WHITE,
    (128, 128, 128): colorama.Fore.LIGHTBLACK_EX,
    (255, 0, 0): colorama.Fore.LIGHTRED_EX,
    (0, 255, 0): colorama.Fore.LIGHTGREEN_EX,
    (255, 255, 0): colorama.Fore.LIGHTYELLOW_EX,
    (0, 255, 255): colorama.Fore.LIGHTCYAN_EX,
    (255, 0, 255): colorama.Fore.LIGHTMAGENTA_EX,
    (255, 255, 255): colorama.Fore.LIGHTWHITE_EX,

}
COLORS = {colorsys.rgb_to_hsv(*k):v for k,v in _COLORS.items()}
def calc_dist(col1,col2):
    return 2*abs(col1[0] - col2[0])+ abs(col1[1]-col2[1])+ abs(col1[2]-col2[2])
def closest_color(color):
    color=colorsys.rgb_to_hsv(*color)
    min_d=min(COLORS,key=lambda d:calc_dist(color,d))
    return COLORS[min_d]
    



try:
    from bs4 import BeautifulSoup as bs
except:
    bs=None
MARKUP = {"b" : ("\x1b[1m","\x1b[22m"),"i": ("\x1b[3m","\x1b[23m"),"u":("\x1b[4m","\x1b[24m"),"s":("\x1b[9m","\x1b[29m"),"em":("\x1b[7m","\x1b[27m")}
ALIAS = {"strong":"b"}
def get_ansi(q):
    if q in ALIAS:
        q=ALIAS[q]
    return MARKUP.get(q,("",""))
def parse(text):
    if bs is None:
        return text
    soup=bs(text,"html.parser")
    out = ""
    for p in soup.contents:
        if isinstance(p,str):
            n=None
            t=p
        else:
            n=p.name
            t=parse(p.encode_contents())
        start,end=get_ansi(n)
        out+=start
        out+=t
        out+=end
    return out
        



from . import srt
from ascvid.logger import print_error
import os

SUB_EXTENSIONS = {".srt":srt}

def Subtitles(t):
    _,f = os.path.splitext(t)
    if f not in SUB_EXTENSIONS:
        print_error(f"Invalid subtitle extension. Implemented extensions: {list(SUB_EXTENSIONS.keys())}. Current extension: {f}")
        exit(1)
    subt_class=SUB_EXTENSIONS[f].Subtitles
    return subt_class(t)  



import re
from ascvid.utils import html2curses as h2c

class SubError(ValueError):
    pass


NUM = re.compile(r"^\d+$")
TIME = re.compile(
    r"^(\d{2})\:(\d{2})\:(\d{2}\,\d{3}) \-\-\> (\d{2})\:(\d{2})\:(\d{2}\,\d{3})$"
)


def __corr(msg="unknown reason", filename="?", line="?"):
    raise SubError(f"{filename}:{line} => corrupted srt file: [{msg}]")


def calculate_timestamp(h, m, s):
    return h * 3600 + m * 60 + s


class SubEntry:
    def __init__(self, index, tstamp, contents):
        self.index = index
        self.stamp = tstamp
        self.contents = h2c.parse(contents)
        self.active = self.stamp.active

class Timestamp:
    def __init__(self, start, end):
        self.start_tup = start
        self.end_tup = end
        self.start_s = calculate_timestamp(*start)
        self.end_s = calculate_timestamp(*end)

    def active(self, time):
        return self.start_s <= time <= self.end_s


class SubParser:
    def __init__(self, stream):
        self.stream = stream
        self.line_counter = 0

    def readline(self):
        self.line_counter += 1
        return self.stream.readline()

    def _corr(self, msg):
        __corr(msg, self.stream.name, self.line_counter)
    def iter_subs(self):
        while True:
            first_line = self.readline()

            if not first_line:
                return
            first_line = first_line.strip()
            if NUM.match(first_line):
                index = int(first_line)
            else:
                self._corr("incorrect index line format: {first_line!r}")

            time_line = self.readline()
            if not time_line:
                self._corr("missing timestamp line")
            time_line = time_line.strip()
            if not TIME.match(time_line):
                self._corr(f"wrong timestamp line format: {time_line!r}")
            t_r = TIME.search(time_line).groups()
            h, m, s, *end = t_r
            s = s.replace(",", ".")
            h, m, s = int(h), int(m), float(s)
            eh, em, es = end
            es = es.replace(",", ".")
            eh, em, es = int(eh), int(em), float(es)
            cont_line = self.readline()
            if not cont_line:
                self._corr("missing content")

            content = cont_line
            while True:
                nline = self.readline()
                if not nline.strip():
                    break
                content+=nline
            timestamp = Timestamp((h, m, s), (eh, em, es))
            yield SubEntry(index, timestamp, content)


class Subtitles:
    def __init__(self, subtitle_file):
        self.file=subtitle_file
        with open(self.file) as fi:
            self.subs=list(SubParser(fi).iter_subs())
        self.current_sub_index=0
        self.max_lines=max([len(q.contents.splitlines()) for q in self.subs])

    def get_sub(self,t):
        if self.current_sub_index==len(self.subs):
            return ""
        sub=self.subs[self.current_sub_index]
        if sub.active(t):
            return sub.contents
        elif sub.stamp.end_s<=t:
            self.current_sub_index+=1
            return self.get_sub(t)
        return ""



from PIL import Image
from math import sqrt
import moviepy.editor as mp
import sys
import os
import time
import cursor
import threading
import getkey
import queue
import imageio
from .getcol import closest_color
from .logger import print_error,print_warning
from . import audio
from . import subs as sus
from .timer import Timer
CHARS = ' .\'`^",:;Il!i><~+_-?][}{1)(|\\/tfjrxnuvczXYUJCLQ0OZmwqpdbkhao*#MW&8%B@$\u2591\u2592\u2593\u2588'
CLOSEST_CACHE = {}
CHAR_CACHE = {}
def fqget(q):
    if q.empty():
        return
    return q.get()
def _getchar(color):
    if color in CHAR_CACHE:
        return CHAR_CACHE[color]
    brightness=round(get_brightness(color))
    charindex=int((brightness/255)*(len(CHARS)-1))
    char=CHARS[charindex]
    CHAR_CACHE[color]=char
    return char

def check_paused(q):
    def check():
        while True:
            q.put(getkey.getkey())
    return check
def closest(col):
    global CLOSEST_CACHE
    if col in CLOSEST_CACHE:
        return CLOSEST_CACHE[col]
    cl=closest_color(col)
    CLOSEST_CACHE[col]=cl

    return cl
def clear(out):
    out.write("\x1b[H\x1b[2J\x1b[3J")
def get_brightness(col):
    R,G,B=col
    return sqrt(0.299 * R**2 + 0.587 * G**2 + 0.114 * B**2)

def get_character(color,colored,truecolor):
    r,g,b=color
    brightness=round(get_brightness(color))
    charindex=int((brightness/255)*(len(CHARS)-1))
    character=CHARS[charindex]
    if colored:
        if truecolor:
            return f"\x1b[38;2;{r};{g};{b}m{character}"
        return f"{closest(color)}{character}"
    return character
def get_pixel(color,colored,truecolor,use_ascii,char):
    if use_ascii:
        return get_character(color,colored,truecolor)
    if truecolor:
        r,g,b=color
        return f"\x1b[38;2;{r};{g};{b}m{char}"
    return f"{closest(color)}{char}"

def show_frame(fr,char,colored,truecolor,use_ascii,resize,hast,tit,out,fps,nfps,deb,subs,ttt,dur,avg):
    ts=os.get_terminal_size()
    tw=ts.columns
    th=ts.lines
    if deb:
        th-=1
   
 
    if hast:
        th-=1
    if subs:
        th-=subs.max_lines

    out.write('\033[H')
    out.flush()
    d=Image.fromarray(fr)
    if resize:
        dh=d.height//2
        wd=d.width/tw
        hd=dh/th
        scale=1
        if wd>1 and hd>1 and wd>=hd:
            scale=tw/d.width
        elif wd>1 and hd>1:
            scale=th/dh
        elif wd>1:
            scale=tw/d.width
        elif hd>1:
            scale=th/dh

        d=d.resize((int(d.width*scale),int(dh*scale)))
    pix=d.load()
    lines=[]

    for y in range(d.height):
        line=''
        for x in range(d.width):
            r,g,b,*foo=pix[x,y]
            line+=get_pixel((r,g,b),colored,truecolor,use_ascii,char)
        lines.append(line+'\x1b[0m')
    clear(out)
    if hast:
        print(tit,file=out,flush=True)
    if deb:
        print(f"{ttt:.2f}/{dur}@ {fps:.2f} FPS ({avg} AVG)  => {nfps} FPS",flush=True,file=out)
    print('\n'.join(lines),end="",file=out)
    if subs:
        print("\n"+subs.get_sub(ttt),end="",file=out)
def mkpos(a):
    if a<0:
        return 0
    return a
def play_vid(file,hide_cursor=True,play_audio=True,fps=None,char="\u2588",colored=True,truecolor=True,use_ascii=False,fast=False,disable_controls=False,title=None,show_title=True,out=None,show_dbg=True,subs=None):
    if out is None:
        out=sys.stdout
    else:
        out=open(out,"w")
    if title is None:
        title=file
    if subs is not None:
        subs=sus.Subtitles(subs)
    

    audio_clip=mp.AudioFileClip(file)
    if len(audio_clip.reader.buffer)==0:
        audio_clip=None
    if fps not in (None,"max"):
        vid=imageio.read(file,fps=fps,)
    else:
        vid=imageio.read(file,)
    fps=vid.get_meta_data().get("fps",mp.VideoFileClip(file).fps)
    q=queue.Queue()
    dur=vid.get_meta_data().get("duration",float("inf"))
    settings=None
    if sys.platform.lower().startswith("linux"):
        import tty
        import termios
        settings=termios.tcgetattr(sys.stdin)
    fpsum=0
    if fast:
        frame_count=vid.count_frames()
        if frame_count>=10000:
            response=print_warning(f"{file!r} has {frame_count} frames.",ask="Are you sure you want to resize them all at once [y*]")
            if response!="y":
                sys.exit(1)
        tw=os.get_terminal_size().columns
        th=os.get_terminal_size().lines
        if show_title:
            th-=1
        if show_dbg:
            th-=1
        if subs is not None:
            th-=subs.max_lines
        vh,vw=vid.get_meta_data()["size"]
        dh=vh//2
        wd=vw-tw
        hd=dh-th
        scale=1
        if wd>0 and hd>0 and wd>=hd:
            scale=tw/vw
        elif wd>0 and hd>0:
            scale=th/dh
        elif wd>0:
            scale=tw/vw
        elif hd>0:
            scale=th/dh

        vid=imageio.read(file,size=(int(vw*scale),int(dh*scale)),fps=fps,)

    process=None
    audio_stream=None
    if hide_cursor:
        cursor.hide()
    if play_audio and audio_clip is not None:
        audio_stream=audio.Audio(audio_clip)
        audio_stream.play()
    if not disable_controls:
        check_thread=threading.Thread(target=check_paused(q),daemon=True)
        check_thread.start()
    clear(out)
    frame_timer=Timer()
    frame_iterator=vid.iter_data()
    frindex=0
    try:
        frame_timer.start()
        while True:
            try:
                frame=next(frame_iterator)
            except StopIteration:
                break

            if not disable_controls:
                resp=fqget(q)
                if resp==" ":
                    frame_timer.pause()
                    if audio_stream is not None:
                        audio_stream.pause()
                    while q.get()!=" ":
                        pass
                    if audio_stream is not None:
                        audio_stream.resume()
                    frame_timer.start()
            current_fps=frindex/frame_timer.curtime
            avgfps=0 if frindex==0 else fpsum/frindex
            show_frame(frame,char,colored,truecolor,use_ascii,not fast,show_title,title,out,current_fps,fps,show_dbg,subs,frame_timer.curtime,dur,avgfps)
            if fps!="max" and current_fps>fps:
                time.sleep(mkpos(1/fps))
            frindex+=1
            fpsum+=current_fps
    except KeyboardInterrupt:
        clear(out)
    finally:
        cursor.show()
        out.write("\x1b[0m")
        out.flush()
        if settings is not None:
            time.sleep(0.2)
            termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, settings)
    clear(out)


           




import setuptools

setuptools.setup(name="ascvid",version="1.5.3",description="ASCII Video player.", long_description=
                 """
.. image:: https://raw.githubusercontent.com/jenca-adam/ascvid/main/asc.png

``ascvid`` is an ASCII video player with quite exact results. It is mostly advised to be used under Linux, but it might work on Windows too (its behavior under Mac is untested, and it will likely not work well).
Here is a little showcase of what it can do: 

.. image:: https://raw.githubusercontent.com/jenca-adam/ascvid/main/rick.gif

It requires a Truecolor terminal to work like this. If you are on a stupider terminal, the results will look less realistic.The produced graphics aren't blinking as in other ASCII video players, however, the videos might lag a bit if your terminal is zoomed out.
``ascvid`` also supports audio output and pausing your video! It's a true video player.


Installation
============

.. code-block:: console
   
   python3 -m pip install ascvid

Then you can just run ``ascvid`` and you're good to go.

CLI Options
===========
*NOTE :: --no-truecolor option lags like my brain, so if you are on a stupid terminal, you are better off using --no-color in ASCII mode.*
.. code-block:: console
   
    Usage: ascvid [OPTIONS] FILE

    Options:
      -H, --hide-cursor       Hide the cursor while playing the video
      -A, --no-audio          Don't play audio stream
      -f, --fps TEXT          Number of FPS the video's supposed to run at. If
                              None, it's determined from the video. If "max",
                              ascvid will try its best to keep the video from
                              lagging
      -c, --char TEXT         Character to be used while rendering the video
                              frames
      -C, --no-color          Don't color output
      -a, --ascii             Use multiple ASCII characters. Best to be used with
                              --no-truecolor
      -T, --no-truecolor      Reduces color palette. Use this flag on more stupid
                              terminals (windows).
      -F, --fast              Toggles off resizing each frame individually, rather
                              resizes the entire video. Use this if the video is
                              lagging too much.
      -d, --disable-controls  Disables pausing the video
      -t, --title TEXT        Sets the title of the video. If not set, file name
                              will be used instead
      -h, --hide-title        hides the title
      -n, --new-window        Opens in a new terminal window
      -t, --term TEXT         Specify terminal in format '<terminal command> <run
                              command switch> {}'
      -o, --outfile TEXT      Specify output file
      -d, --debug             Debug mode: show current/target fps
      -s, --subs              Subtitle file to use

      --help                  Show this message and exit.
   
    """,
                 install_requires=["pillow","moviepy","cursor","numpy","click","sounddevice","av","imageio"],
                 packages=["ascvid","ascvid.subs","ascvid.utils","ascvid.subs.srt"],
                 classifiers=["Development Status :: 4 - Beta","Environment :: Console","Intended Audience :: End Users/Desktop","License :: OSI Approved :: GNU General Public License v3 (GPLv3)","Operating System :: Microsoft :: Windows","Operating System :: POSIX :: Linux","Operating System :: Unix","Programming Language :: Python :: 3.10","Topic :: Multimedia :: Video :: Display"],keywords="Video,ASCII,Terminal",author="Adam Jenca",author_email="jenca.a@gjh.sk",url="https://github.com/jenca-adam/ascvid",entry_points={"console_scripts":["ascvid = ascvid.cli:main"]})









import subprocess
import shutil
import sys
import shlex
from .logger import print_error
UNIX_XTERM_SYNTAX = ["konsole","rxvt","uxterm","xterm -bg black -fg white"]
def get_term_cmd():
    if sys.platform=="win32":
        return "start \'\' \'{}\'"
    elif sys.platform=="darwin":
        return "open -W -a Terminal.app \'{}\'"
    elif sys.platform.lower().startswith("linux"):
        if shutil.which("gnome-terminal"):
            return "gnome-terminal -x \'{}\'"
        for term in UNIX_XTERM_SYNTAX:
            if shutil.which(term):
                return f"{term} -e \'{{}}\'"

def run_term(cmd,term_cmd=None):
    if term_cmd is None:
        term_cmd=get_term_cmd()
        if term_cmd is None:
            print_error(f"We could not find a terminal for you! Current terminal list: {UNIX_XTERM_SYNTAX}. If your terminal is not there, specify the terminal you are using with --term option. \n If you think this is an error with ascvid, please use the GitHub issue tracker.")
            exit(1)
    term=term_cmd.format(cmd)
    print(shlex.split(term))
    popen=subprocess.Popen(shlex.split(term),stderr=subprocess.PIPE)
    stderr_out,*_=popen.communicate()
    if popen.returncode!=0:
        print_error(f"Command {term} failed with returncode {popen.returncode}! STDERR output: \n{stderr_out}")



from .player import play_vid
from .run_terminal import run_term 
import click
import sys
@click.command()
@click.argument("file")
@click.option("--hide-cursor","-H",default=False,is_flag=True,help="Hide the cursor while playing the video")
@click.option("--no-audio","-A",is_flag=True,default=False,help="Don't play audio stream")
@click.option("--fps","-f",default=None,help="Number of FPS the video's supposed to run at. If None, it's determined from the video. If \"max\", ascvid will try its best to keep the video from lagging")
@click.option("--char","-c",default='\u2588',type=str,help="Character to be used while rendering the video frames")
@click.option("--no-color","-C",default=False,is_flag=True,help="Don't color output")
@click.option("--ascii","-a",is_flag=True,default=False,help="Use multiple ASCII characters. Best to be used with --no-truecolor")
@click.option("--no-truecolor","-T",is_flag=True,default=False,help="Reduces color palette. Use this flag on more stupid terminals (windows).")
@click.option("--fast","-F",is_flag=True,default=False,help="Toggles off resizing each frame individually, rather resizes the entire video. Use this if the video is lagging too much.")
@click.option("--disable-controls","-d",is_flag=True,default=False,help="Disables pausing the video")
@click.option("--title","-t",default=None,help="Sets the title of the video. If not set, file name will be used instead")
@click.option("--hide-title","-h",default=False,is_flag=True,help="hides the title")
@click.option("--new-window","-n",is_flag=True,help="Opens in a new terminal window")
@click.option("--term","-t",default=None,help="Specify terminal in format '<terminal command> <run command switch> {}'")
@click.option("--outfile","-o",default=None,help="Specify output file")
@click.option("--debug","-d",is_flag=True,help="Debug mode: show current/target FPS")
@click.option("--subs","-s",default=None,help="subtitle file to use")
def main(file,hide_cursor,no_audio,fps,char,no_color,ascii,no_truecolor,fast,disable_controls,title,hide_title,new_window,term,outfile,debug,subs):
    if new_window:
        cmd = [sys.executable,"-m","ascvid",file]
        if hide_cursor:
            cmd.extend(["-H"])
        if no_audio:
            cmd.extend(["-A"])
        if fps is not None:
            cmd.extend(["-f"])
            cmd.extend([f"{fps}"])
        cmd.extend(["-c",f"{char}"])
        if no_color:
            cmd.extend(["-C"])
        if ascii:
            cmd.extend(["-a"])
        if no_truecolor:
            cmd.extend(["-T"])
        if fast:
            cmd.extend(["-F"])
        if disable_controls:
            cmd.extend(["-d"])
        if title is not None:
            cmd.extend(["-t",f"{title}"])
        if hide_title:
            cmd.extend(["-h"])
        if debug:
            cmd.extend(["-d"])
        run_term(' '.join(cmd),term)
        return
    if fps and fps!="max":
        fps=int(fps)
    play_audio=not no_audio
    colored = not no_color
    truecolor = not no_truecolor
    play_vid(file,hide_cursor,play_audio,fps,char,colored,truecolor,ascii,fast,disable_controls,title,not hide_title,outfile,debug,subs)
